/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.impl;

import org.xmloperator.lambda.net.model.Applicator;
import org.xmloperator.lambda.net.model.Delimiter;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.PortStack;

/**
 * Lambda-net branch pruning.
 */
public abstract class Pruner {

	/**
	 * Return to the factory the branch of the lambda-net starting at a given Port.
	 * A Delimiter is returned to factory if its input Port or its body Port
	 *   is disconnected or connected to an Eraser.
	 * An Abstractor, an Applicator or a Duplicator is returned to factory if
	 *   its principal Ports is disconnected (or connected to an Eraser)
	 *   or if its two auxiliary Ports are disconnected (or connected to Erasers).
	 * Free ports are reconnected to Erasers.
	 * Pre-condition: the given starting Port is disconnected.
	 * @param startingPort a Port from which the process begins.
	 * @param canReturnAnApplicatorForOnlyOneErasedPort if true then an applicator
	 *   that has only one erased Port will be returned to factory.
	 * @throws an IllegalArgumentException if the given starting Port is connected.
	 */
	public static final void returnBranchToFactory(
		final Port startingPort,
		boolean canReturnAnApplicatorForOnlyOneErasedPort) {
		if (startingPort.getConnectedPort() != null)
			throw new IllegalArgumentException();
		final Operator startingOperator = startingPort.getOperator();
		final int mark = startingOperator.getMark();
		final OperatorFactory operatorFactory = startingOperator.getFactory();
		final PortStack portStack = operatorFactory.createPortStack();
		try {
			portStack.push(startingPort);
			while (!portStack.isEmpty()) {
				Port port = portStack.pop();
				Operator operator = port.getOperator();
				if (operator == null || operator.isReturnedToFactory());
				else if (operator.isEraser())
					operator.returnToFactory();
				else if (operator.isDelimiter()) {
					if (!port.isSecondAuxiliary()) {
						Delimiter delimiter = (Delimiter) operator;
						{
							Port input = delimiter.getInput().getConnectedPort();
							if (input != null) {
								input.disconnect();
								portStack.push(input);
							}
						}
						{
							Port body = delimiter.getBody().getConnectedPort();
							if (body != null) {
								body.disconnect();
								portStack.push(body);
							}
						}
						delimiter.returnToFactory();
					}
				}
				else if (
					canReturnAnApplicatorForOnlyOneErasedPort
						&& operator.isApplicator()) {
					{
						Applicator applicator = (Applicator) operator;
						Port input = applicator.getInput().getConnectedPort();
						Port body = applicator.getBody().getConnectedPort();
						Port argument = applicator.getArgument().getConnectedPort();
						if (input != null) {
							input.disconnect();
							portStack.push(input);
						}
						if (body != null) {
							body.disconnect();
							portStack.push(body);
						}
						if (argument != null) {
							argument.disconnect();
							portStack.push(argument);
						}
						applicator.returnToFactory();
					}
				}
				else {
					Port port1 = operator.getPort1();
					Port port2 = operator.getPort2();
					Port port1c = port1.getConnectedPort();
					Port port2c = port2.getConnectedPort();
					if (port.isPrincipal()
						|| ((port1c == null || port1c.getOperator().isEraser())
							&& (port2c == null || port2c.getOperator().isEraser()))) {
						{
							Port port0c = operator.getPort0().getConnectedPort();
							if (port0c != null) {
								port0c.disconnect();
								portStack.push(port0c);
							}
						}
						if (port1c != null) {
							port1c.disconnect();
							portStack.push(port1c);
						}
						if (port2c != null) {
							port2c.disconnect();
							portStack.push(port2c);
						}
						operator.returnToFactory();
					}
					else {
						Eraser eraser =
							operatorFactory.createEraser(mark, !port.isInput(), true);
						port.connectTo(eraser.getPort());
					}
				}
			}
		}
		finally {
			portStack.returnToFactory();
		}
	}
}
