/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.util;

import org.xmloperator.lambda.net.exception.FreePortException;
import org.xmloperator.lambda.net.exception.MarkedDifferentlyException;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.PortStack;

/**
 * This is an iterator on the Operators belonging to a lambda-net.
 * <p>The mark of each Operator is changed during traversal.</p>
 * <p>The net is supposed to be invariant (no interaction) during traversal.</p>
 * <p>The first Operator given by the <tt>next()</tt> method is the root Eraser.</p>
 * <p>Before traversing a net, its Operators must have the same mark.</p>
 */
public class NetTraversal implements OperatorIterator {
	final int oldMark;
	final int newMark;
	private PortStack portStack;
	private Operator operator;

	public NetTraversal(Eraser rootEraser) {
		this(rootEraser, true);
	}

	public NetTraversal(Eraser rootEraser, boolean isMarkIncremented) {
		this.oldMark = rootEraser.getMark();
		this.newMark = this.oldMark + (isMarkIncremented ? 1 : -1);
		OperatorFactory operatorFactory = rootEraser.getFactory();
		this.portStack = operatorFactory.createPortStack();
		this.portStack.push(rootEraser.getPort());
		this.operator = rootEraser;
		rootEraser.setMark(this.newMark);
	}

	private void nextOperator() {
		while (true) {
			if (this.portStack.isEmpty()) {
				this.operator = null;
				this.portStack.returnToFactory();
				this.portStack = null;
				return;
			}
			Port port = this.portStack.pop();
			Port connectedPort = port.getConnectedPort();
			if (connectedPort == null)
				throw new FreePortException(port);
			Operator operator = connectedPort.getOperator();
			int mark = operator.getMark();
			if (mark == this.newMark)
				continue;
			if (mark != this.oldMark)
				throw new MarkedDifferentlyException(operator);
			operator.setMark(this.newMark);
			this.operator = operator;
			switch (operator.getPortCount()) {
				case 3 :
					{
						Port port2 = operator.getPort2();
						if (port2 != connectedPort)
							portStack.push(port2);
					}
				case 2 :
					{
						Port port1 = operator.getPort1();
						if (port1 != connectedPort)
							portStack.push(port1);
					}
					{
						Port port0 = operator.getPort0();
						if (port0 != connectedPort)
							portStack.push(port0);
					}
			}
			return;
		}
	}

	//////////////////////////////////
	// OperatorIterator implementation
	//////////////////////////////////

	public boolean hasNext() {
		return this.operator != null;
	}

	public Operator next() {
		Operator operator = this.operator;
		if (operator == null)
			throw new IllegalStateException();
		this.nextOperator();
		return operator;
	}
}
