/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.shared;

/**
 * Stack using an ObjectArrayFactory.
 */
public class BaseStack {
	static final int INITIAL_LENGTH = 8;
	protected final ObjectArrayFactory objectArrayFactory;
	public boolean acceptNull;
	protected Object[] objects = null;
	protected int size = 0;

	public BaseStack(ObjectArrayFactory objectArrayFactory, boolean acceptNull) {
		this.objectArrayFactory = objectArrayFactory;
		this.acceptNull = acceptNull;
	}

	public void startUse() {
		if (this.objects != null)
			throw new IllegalStateException();
		this.objects =
			this.objectArrayFactory != null
				? this.objectArrayFactory.createObjectArray(INITIAL_LENGTH)
				: new Object[INITIAL_LENGTH];
	}

	public void endUse() {
		if (this.objects == null)
			throw new IllegalStateException();
		this.size = 0;
		if (this.objectArrayFactory != null)
			this.objectArrayFactory.returnObjectArray(this.objects);
		this.objects = null;
	}

	public int getSize() {
		return this.size;
	}

	public boolean isEmpty() {
		return this.size == 0;
	}

	public void clear() {
		for (int k = 0; k < this.size; k++)
			this.objects[k] = null;
		this.size = 0;
	}

	void ensureOneMore() {
		if (this.objects.length == this.size) {
			int newLength = 2 * objects.length;
			Object[] newObjects =
				objectArrayFactory != null
					? objectArrayFactory.createObjectArray(newLength)
					: new Object[newLength];
			System.arraycopy(this.objects, 0, newObjects, 0, this.size);
			if (objectArrayFactory != null)
				objectArrayFactory.returnObjectArray(this.objects);
			this.objects = newObjects;
		}
	}

	protected void pushObject(Object object) {
		if (object == null && !this.acceptNull)
			throw new IllegalArgumentException();
		this.ensureOneMore();
		this.objects[this.size++] = object;
	}

	protected Object popObject() {
		if (this.size == 0)
			throw new IllegalStateException();
		int index = --this.size;
		Object object = this.objects[index];
		this.objects[index] = null;
		return object;
	}

	protected Object topObject() {
		return this.size > 0 ? this.objects[this.size - 1] : null;
	}

	protected boolean containsObject(Object object) {
		for (int k = this.size - 1; k >= 0; k--)
			if (this.objects[k] == object)
				return true;
		return false;
	}

	protected boolean removeObject(Object object) {
		int index = -1;
		for (int k = this.size - 1; k >= 0; k--)
			if (this.objects[k] == object) {
				index = k;
				break;
			}
		if (index < 0)
			return false;
		int size = --this.size;
		System.arraycopy(
			this.objects,
			index + 1,
			this.objects,
			index,
			size - index);
		this.objects[size] = null;
		return true;
	}

	protected void insertObjectFromTop(int index, Object object) {
		if (object == null)
			throw new IllegalArgumentException();
		this.ensureOneMore();
		int k = this.size++ -index;
		System.arraycopy(this.objects, k, this.objects, k + 1, index);
		this.objects[k] = object;
	}

	protected Object removeObjectFromTop(int index) {
		int size = --this.size;
		int k = size - index;
		Object object = this.objects[k];
		System.arraycopy(this.objects, k + 1, this.objects, k, index);
		this.objects[size] = null;
		return object;
	}
}
