/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.translate;

import org.xmloperator.lambda.net.impl.Pruner;
import org.xmloperator.lambda.net.model.Abstractor;
import org.xmloperator.lambda.net.model.Applicator;
import org.xmloperator.lambda.net.model.Delimiter;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.PortStack;
import org.xmloperator.lambda.net.util.InteractionListener;
import org.xmloperator.lambda.net.util.NetTraversal;
import org.xmloperator.lambda.net.util.NetUtils;
import org.xmloperator.lambda.net.util.OperatorIterator;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.model.TermWithBody;

/**
 * Translation from lambda-net to Term.
 */
public abstract class FromNetToTerm {
	private static final boolean SCOPE_BINDING = Test.SCOPE_BINDING;
	private static final boolean DEBUG = false;

	/**
	 * Translate an interaction net to a Term.
	 * <p><b>Pre-conditions:</b></p><ul>
	 * <li>there is no reducable interaction other than beta and the root Eraser.</li>
	 * <li>the root Eraser is inactive.</li></ul>
	 * <p><b>Post-condition:</b> the net is consumed.</p>
	 * @param rootEraser a root Eraser.
	 * @param interactionListener an InteractionListener.
	 * @return the resulting Term.
	 */
	public static final Term translate(
		final Eraser rootEraser,
		TermFactory termFactory,
		InteractionListener interactionListener) {
		if (SCOPE_BINDING)
			removeScopeBinding(rootEraser);
		if (DEBUG) {
			System.out.println("After removing scope binding");
			NetUtils.print(System.out, rootEraser);
		}
		unwind(rootEraser, interactionListener);
		if (DEBUG) {
			System.out.println("After unwind");
			NetUtils.print(System.out, rootEraser);
		}
		removeScope(rootEraser, interactionListener);
		if (DEBUG) {
			System.out.println("After removing scope");
			NetUtils.print(System.out, rootEraser);
		}
		cutLoop(rootEraser, interactionListener);
		if (DEBUG) {
			System.out.println("After cutting the loop");
			NetUtils.print(System.out, rootEraser);
		}
		Term term = translateTree(rootEraser, termFactory);
		returnNetToFactory(rootEraser);
		return term;
	}

	/**
	 * Returns a lambda-net to the factory.
	 * @param rootEraser a root Eraser.
	 */
	public static final void returnNetToFactory(final Eraser rootEraser) {
		Port port = rootEraser.getPort().getConnectedPort();
		port.disconnect();
		Pruner.returnBranchToFactory(port, true);
		rootEraser.returnToFactory();
	}

	private static final void apply(
		PortStack portStack,
		InteractionListener interactionListener) {
		while (!portStack.isEmpty()) {
			Operator operator = portStack.pop().getOperator();
			if (!operator.isReturnedToFactory()) {
				int interactionType = operator.interact(portStack);
				if (interactionListener != null)
					interactionListener.notifyInteraction(interactionType);
			}
		}
	}

	private static final void removeScopeBinding(Eraser rootEraser) {
		OperatorFactory operatorFactory = rootEraser.getFactory();
		PortStack portStack = operatorFactory.createPortStack();
		try {
			{
				OperatorIterator iterator = new NetTraversal(rootEraser);
				while (iterator.hasNext()) {
					Operator operator = iterator.next();
					if (operator.isAbstractor())
						portStack.push(operator.getPort0());
				}
			}
			while (!portStack.isEmpty()) {
				Abstractor abstractor = (Abstractor) portStack.pop().getOperator();
				abstractor.detachScopeBinder();
			}
		}
		finally {
			portStack.returnToFactory();
		}
	}

	private static final void unwind(
		Eraser rootEraser,
		InteractionListener interactionListener) {
		{
			OperatorIterator iterator = new NetTraversal(rootEraser);
			while (iterator.hasNext()) {
				Operator operator = iterator.next();
				if (operator.isApplicator())
					 ((Applicator) operator).reverseOrientation();
			}
		}
		PortStack portStack = rootEraser.getFactory().createPortStack();
		try {
			OperatorIterator iterator = new NetTraversal(rootEraser);
			while (iterator.hasNext()) {
				Operator operator = iterator.next();
				if (operator.isApplicator()) {
					Port principalPort = operator.getPort0();
					if (principalPort.getConnectedPort().isPrincipal())
						portStack.push(principalPort);
				}
			}
			apply(portStack, interactionListener);
		}
		finally {
			portStack.returnToFactory();
		}
	}

	private static final void removeScope(
		Eraser rootEraser,
		InteractionListener interactionListener) {
		{
			OperatorIterator iterator = new NetTraversal(rootEraser);
			while (iterator.hasNext()) {
				Operator operator = iterator.next();
				if (operator.isDelimiter())
					 ((Delimiter) operator).reverseOrientation();
			}
		}
		PortStack portStack = rootEraser.getFactory().createPortStack();
		try {
			OperatorIterator iterator = new NetTraversal(rootEraser);
			while (iterator.hasNext()) {
				Operator operator = iterator.next();
				if (operator.isDelimiter()) {
					Port principalPort = operator.getPort0();
					if (principalPort.getConnectedPort().isPrincipal())
						portStack.push(principalPort);
				}
			}
			apply(portStack, interactionListener);
		}
		finally {
			portStack.returnToFactory();
		}
	}

	private static final void cutLoop(
		Eraser rootEraser,
		InteractionListener interactionListener) {
		OperatorFactory operatorFactory = rootEraser.getFactory();
		PortStack bindPortStack = operatorFactory.createPortStack();
		PortStack portStack = operatorFactory.createPortStack();
		try {
			{
				OperatorIterator iterator = new NetTraversal(rootEraser);
				while (iterator.hasNext()) {
					Operator operator = iterator.next();
					if (operator.isAbstractor())
						bindPortStack.push(((Abstractor) operator).getBind());
				}
			}
			int mark = rootEraser.getMark();
			while (!bindPortStack.isEmpty()) {
				Port bindPort = bindPortStack.pop();
				Port connectedPort = bindPort.getConnectedPort();
				bindPort.disconnect();
				bindPort.connectTo(
					operatorFactory.createEraser(mark, false, true).getPort());
				Port eraserPort =
					operatorFactory.createEraser(mark, true, true).getPort();
				eraserPort.connectTo(connectedPort);
				portStack.notifyConnection(eraserPort, connectedPort);
			}
			apply(portStack, interactionListener);
		}
		finally {
			bindPortStack.returnToFactory();
			portStack.returnToFactory();
		}
	}

	private static final Term translateTree(
		final Eraser rootEraser,
		TermFactory termFactory) {
		Term rootTerm = null;
		PortStack portStack = rootEraser.getFactory().createPortStack();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Operator operator = rootEraser.getPort().getConnectedPort().getOperator();
			Term parent = null;
			while (true) {
				Term term;
				if (operator.isEraser())
					term = termFactory.createLeaf();
				else if (operator.isDelimiter())
					term = termFactory.createEndOfScope();
				else if (operator.isAbstractor())
					term = termFactory.createAbstraction();
				else if (operator.isApplicator())
					term = termFactory.createApplication();
				else
					throw new IllegalArgumentException();
				if (parent == null)
					rootTerm = term;
				else if (parent.isApplication()) {
					Application application = (Application) parent;
					if (application.getBody() == null)
						application.setBody(term);
					else
						application.setArgument(term);
				}
				else
					 ((TermWithBody) parent).setBody(term);
				if (!operator.isEraser()) {
					if (operator.isApplicator()) {
						portStack.push(
							((Applicator) operator).getArgument().getConnectedPort());
						termStack.push(term);
					}
					operator = operator.getPort1().getConnectedPort().getOperator();
					parent = term;
					continue;
				}
				if (portStack.isEmpty())
					return rootTerm;
				operator = portStack.pop().getOperator();
				parent = termStack.pop();
			}
		}
		finally {
			portStack.returnToFactory();
			termStack.returnToFactory();
		}
	}
}
