/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.translate;

import org.xmloperator.lambda.net.model.Abstractor;
import org.xmloperator.lambda.net.model.Applicator;
import org.xmloperator.lambda.net.model.Delimiter;
import org.xmloperator.lambda.net.model.Duplicator;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.PortStack;
import org.xmloperator.lambda.net.model.ScopeBinder;
import org.xmloperator.lambda.net.util.NetTraversal;
import org.xmloperator.lambda.net.util.NetUtils;
import org.xmloperator.lambda.net.util.OperatorIterator;
import org.xmloperator.lambda.tree.exception.MissingContextException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermStack;

/**
 * Translation of a Term into a lambda-net.
 */
public abstract class FromTermToNet {
	private static final boolean ASSERTION = Test.ASSERTION;
	private static final boolean SCOPE_BINDING = Test.SCOPE_BINDING;

	/**
	 * Translate a closed Term to a lambda-net.
	 * <p><b>Post-condition:</b> all the Operators have the same mark.</p>
	 * <p><b>Post-condition:</b> there is no reducable interaction other than beta
	 *   and the root Eraser.</p>
	 * @param rootTerm a Term to translate.
	 * @param rootEraser a starting root Eraser, whose mark will be copied.
	 * @throws MissingContextException if the given Term is not closed.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final void translate(
		final Term rootTerm,
		final Eraser rootEraser) {
		OperatorFactory operatorFactory = rootEraser.getFactory();
		PortStack outputPortStack = operatorFactory.createPortStack();
		PortStack outputPortUnStack = operatorFactory.createPortStack();
		PortStack abstractorPortStack =
			SCOPE_BINDING ? operatorFactory.createPortStack() : null;
		PortStack abstractorPortUnStack =
			SCOPE_BINDING ? operatorFactory.createPortStack() : null;
		TermStack termStack = rootTerm.getFactory().createTermStack(false);
		try {
			Term term = rootTerm;
			Port inputPort = rootEraser.getPort();
			int mark = rootEraser.getMark();
			while (true) {
				if (term.isLeaf()) {
					if (outputPortStack.isEmpty())
						throw new MissingContextException(term);
					inputPort.connectTo(outputPortStack.top());
					for (int k = 0; k < outputPortStack.getSize() - 1; k++) {
						Eraser eraser = operatorFactory.createEraser(mark, false, true);
						eraser.getPort().connectTo(outputPortStack.get(k));
					}
				}
				else if (term.isEndOfScope()) {
					{
						ScopeBinder scopeBinder = null;
						if (SCOPE_BINDING) {
							Port abstractorPort = abstractorPortStack.pop();
							abstractorPortUnStack.push(abstractorPort);
							Abstractor abstractor = (Abstractor) abstractorPort.getOperator();
							scopeBinder = abstractor.attachScopeBinder();
						}
						Delimiter delimiter =
							operatorFactory.createDelimiter(mark, false, scopeBinder);
						inputPort.connectTo(delimiter.getInput());
						inputPort = delimiter.getBody();
					}
					{
						Port outputPort = outputPortStack.pop();
						outputPortUnStack.push(outputPort);
						Eraser eraser = operatorFactory.createEraser(mark, false, true);
						eraser.getPort().connectTo(outputPort);
					}
					{
						EndOfScope endOfScope = (EndOfScope) term;
						termStack.push(endOfScope);
						term = endOfScope.getBody();
					}
					continue;
				}
				else if (term.isAbstraction()) {
					{
						Abstractor abstractor =
							operatorFactory.createAbstractor(mark, null);
						inputPort.connectTo(abstractor.getInput());
						inputPort = abstractor.getBody();
						outputPortStack.push(abstractor.getBind());
						if (SCOPE_BINDING)
							abstractorPortStack.push(inputPort);
					}
					{
						Abstraction abstraction = (Abstraction) term;
						termStack.push(abstraction);
						term = abstraction.getBody();
					}
					continue;
				}
				else if (term.isApplication()) {
					Applicator applicator = operatorFactory.createApplicator(mark);
					inputPort.connectTo(applicator.getInput());
					inputPort = applicator.getBody();
					for (int k = 0; k < outputPortStack.getSize(); k++) {
						Port outputPort = outputPortStack.get(k);
						Duplicator duplicator =
							operatorFactory.createDuplicator(mark, false);
						duplicator.getOutput().connectTo(outputPort);
						outputPortStack.replace(k, duplicator.getInput1());
					}
					Application application = (Application) term;
					termStack.push(application);
					term = application.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty()) {
						term = null;
						break;
					}
					Term child = term;
					term = termStack.pop();
					if (term.isEndOfScope()) {
						inputPort =
							((Delimiter) inputPort.getOperator())
								.getInput()
								.getConnectedPort();
						outputPortStack.push(outputPortUnStack.pop());
						if (SCOPE_BINDING)
							abstractorPortStack.push(abstractorPortUnStack.pop());
					}
					else if (term.isAbstraction()) {
						inputPort =
							((Abstractor) inputPort.getOperator())
								.getInput()
								.getConnectedPort();
						outputPortStack.pop();
						if (SCOPE_BINDING)
							abstractorPortStack.pop();
					}
					else if (term.isApplication()) {
						Applicator applicator = (Applicator) inputPort.getOperator();
						Application application = (Application) term;
						if (child == application.getBody()) {
							inputPort = applicator.getArgument();
							for (int k = 0; k < outputPortStack.getSize(); k++) {
								Port outputPort = outputPortStack.get(k);
								Duplicator duplicator = (Duplicator) outputPort.getOperator();
								outputPortStack.replace(k, duplicator.getInput2());
							}
							termStack.push(application);
							term = application.getArgument();
							break;
						}
						inputPort = applicator.getInput().getConnectedPort();
						for (int k = 0; k < outputPortStack.getSize(); k++) {
							Port outputPort = outputPortStack.get(k);
							Duplicator duplicator = (Duplicator) outputPort.getOperator();
							outputPortStack.replace(
								k,
								duplicator.getOutput().getConnectedPort());
						}
					}
					else
						throw new IllegalStateException();
				}
				if (term == null)
					break;
			}
		}
		finally {
			outputPortStack.returnToFactory();
			outputPortUnStack.returnToFactory();
			if (SCOPE_BINDING) {
				abstractorPortStack.returnToFactory();
				abstractorPortUnStack.returnToFactory();
			}
			termStack.returnToFactory();
		}
		if (ASSERTION)
			NetUtils.verifyClosure(rootEraser);
		removeNullDuplicators(rootEraser);
		removeEdgeDuplicators(rootEraser);
		if (ASSERTION)
			NetUtils.verifyXFree(rootEraser);
	}

	/**
	 * Removes the Duplicators whose both inputs are connected to Erasers.
	 * @param rootEraser a root Eraser.
	 */
	private static final void removeNullDuplicators(Eraser rootEraser) {
		OperatorFactory operatorFactory = rootEraser.getFactory();
		PortStack portStack = operatorFactory.createPortStack();
		try {
			OperatorIterator iterator = new NetTraversal(rootEraser);
			while (iterator.hasNext()) {
				Operator operator = iterator.next();
				if (operator.isDuplicator()) {
					Duplicator duplicator = (Duplicator) operator;
					if (isNullDuplicator(duplicator))
						portStack.push(duplicator.getOutput());
				}
			}
			while (!portStack.isEmpty()) {
				Port outputPort = portStack.pop();
				Duplicator duplicator = (Duplicator) outputPort.getOperator();
				Port eraserPort1 = duplicator.getInput1().getConnectedPort();
				Port eraserPort2 = duplicator.getInput2().getConnectedPort();
				Port outputConnectedPort = outputPort.getConnectedPort();
				eraserPort1.disconnect();
				eraserPort2.disconnect();
				outputConnectedPort.disconnect();
				duplicator.returnToFactory();
				eraserPort1.getOperator().returnToFactory();
				eraserPort2.connectTo(outputConnectedPort);
				Operator connectedOperator = outputConnectedPort.getOperator();
				if (connectedOperator.isDuplicator()) {
					Duplicator connectedDuplicator = (Duplicator) connectedOperator;
					if (isNullDuplicator(connectedDuplicator))
						portStack.push(connectedDuplicator.getOutput());
				}
			}
		}
		finally {
			portStack.returnToFactory();
		}
	}

	private static final boolean isNullDuplicator(Duplicator duplicator) {
		return duplicator.getInput1().getConnectedPort().getOperator().isEraser()
			&& duplicator.getInput2().getConnectedPort().getOperator().isEraser();
	}

	/**
	 * Removes the Duplicators whose one input is connected to an Eraser.
	 * @param rootEraser a root Eraser.
	 */
	private static final void removeEdgeDuplicators(Eraser rootEraser) {
		OperatorFactory operatorFactory = rootEraser.getFactory();
		PortStack portStack = operatorFactory.createPortStack();
		try {
			OperatorIterator iterator = new NetTraversal(rootEraser);
			while (iterator.hasNext()) {
				Operator operator = iterator.next();
				if (operator.isDuplicator()) {
					Duplicator duplicator = (Duplicator) operator;
					if (duplicator
						.getInput1()
						.getConnectedPort()
						.getOperator()
						.isEraser())
						portStack.push(duplicator.getInput2());
					else if (
						duplicator.getInput2().getConnectedPort().getOperator().isEraser())
						portStack.push(duplicator.getInput1());
				}
			}
			while (!portStack.isEmpty()) {
				Port inputPort = portStack.pop();
				Duplicator duplicator = (Duplicator) inputPort.getOperator();
				Port connectedInputPort = inputPort.getConnectedPort();
				Port eraserPort =
					(inputPort.getNumber() == 1
						? duplicator.getInput2()
						: duplicator.getInput1())
						.getConnectedPort();
				Port connectedOutputPort = duplicator.getOutput().getConnectedPort();
				connectedInputPort.disconnect();
				eraserPort.disconnect();
				connectedOutputPort.disconnect();
				duplicator.returnToFactory();
				eraserPort.getOperator().returnToFactory();
				connectedInputPort.connectTo(connectedOutputPort);
			}
		}
		finally {
			portStack.returnToFactory();
		}
	}
}
