/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.translate;

import java.io.PrintStream;

import org.xmloperator.lambda.net.exception.BadFormedException;
import org.xmloperator.lambda.net.model.Abstractor;
import org.xmloperator.lambda.net.model.Applicator;
import org.xmloperator.lambda.net.model.Delimiter;
import org.xmloperator.lambda.net.model.Duplicator;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.IndexedOperator;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.Walk;
import org.xmloperator.lambda.net.model.WalkStack;
import org.xmloperator.lambda.tree.exception.IncorrectTermException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.Term;

/**
 * WalkingContext implementation.
 */
public class WalkingContextTranslaterImpl implements WalkingContextTranslater {
	private final Eraser rootEraser;
	private final OperatorFactory operatorFactory;
	private int bodyIndex = 0;
	private WalkStack walkStack = null;
	private WalkStack walkUnStack = null;
	private Port currentPort;
	private boolean isWalkingForward = true;
	private Term rootTerm = null;
	private Term currentTerm = null;
	private Term nextTerm = null;

	public WalkingContextTranslaterImpl(Eraser rootEraser) {
		this.rootEraser = rootEraser;
		this.operatorFactory = rootEraser.getFactory();
		this.currentPort = rootEraser.getPort();
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		Port port = this.currentPort;
		Operator operator = port.getOperator();
		sb.append(this.isWalkingForward ? '>' : '<');
		sb.append(
			operator.isEraser()
				? "Eraser    "
				: operator.isDelimiter()
				? "Delimit"
				: operator.isAbstractor()
				? "Abstractor"
				: operator.isApplicator()
				? "Applicator"
				: operator.isDuplicator()
				? "Duplica"
				: "???");
		if (operator.isDelimiter() || operator.isDuplicator()) {
			sb.append('[');
			sb.append(((IndexedOperator) operator).getIndex());
			sb.append(']');
		}
		sb.append('.');
		sb.append(port.getNumber());
		sb.append("  ");
		sb.append(bodyIndex);
		if (this.walkStack != null)
			sb.append(this.walkStack);
		return sb.toString();
	}

	////////////////////////////////
	// WalkingContext implementation
	////////////////////////////////

	public Eraser getRootEraser() {
		return this.rootEraser;
	}

	public Port getCurrentPort() {
		return this.currentPort;
	}

	public int getEndOfScopeCount() {
		if (this.currentPort.getNumber() != 2
			|| !this.currentPort.getOperator().isAbstractor())
			throw new IllegalStateException();
		return this.bodyIndex - this.walkStack.top().getBindIndex();
	}

	public boolean isWalkingForward() {
		return this.isWalkingForward;
	}

	public void executeStep() {
		if (this.walkStack == null) {
			this.walkStack = this.operatorFactory.createWalkStack();
			this.walkUnStack = this.operatorFactory.createWalkStack();
		}
		this.currentTerm = this.nextTerm;
		this.currentPort = this.currentPort.getConnectedPort();
		Operator operator = this.currentPort.getOperator();
		if (operator.isApplicator()) {
			Applicator applicator = (Applicator) operator;
			switch (this.currentPort.getNumber()) {
				case 0 : // Body
					if (this.currentTerm != null) {
						if (!this.currentTerm.isApplication())
							throw new IncorrectTermException(this.currentTerm);
						this.nextTerm = ((Application) this.currentTerm).getArgument();
					}
					this.currentPort = applicator.getArgument();
					this.isWalkingForward = true;
					break;
				case 1 : // Input
					if (this.currentTerm != null) {
						if (!this.currentTerm.isApplication())
							throw new IncorrectTermException(this.currentTerm);
						this.nextTerm = ((Application) this.currentTerm).getBody();
					}
					this.currentPort = applicator.getBody();
					break;
				case 2 : // Argument
					if (this.currentTerm != null) {
						if (!this.currentTerm.isApplication())
							throw new IncorrectTermException(this.currentTerm);
						this.nextTerm = this.currentTerm.getParent();
					}
					this.currentPort = applicator.getInput();
					break;
				default :
					throw new IllegalStateException();
			}
		}
		else if (operator.isAbstractor()) {
			switch (this.currentPort.getNumber()) {
				case 0 : // Input
					if (this.currentTerm != null) {
						if (!this.currentTerm.isAbstraction())
							throw new IncorrectTermException(this.currentTerm);
						this.nextTerm = ((Abstraction) this.currentTerm).getBody();
					}
					this.walkStack.push(
						this.operatorFactory.createWalk(++this.bodyIndex));
					this.currentPort = ((Abstractor) operator).getBody();
					break;
				case 1 : // Body
					if (this.currentTerm != null) {
						if (!this.currentTerm.isAbstraction())
							throw new IncorrectTermException(this.currentTerm);
						this.nextTerm = this.currentTerm.getParent();
					}
					this.walkStack.pop().returnToFactory();
					this.bodyIndex--;
					this.currentPort = ((Abstractor) operator).getInput();
					break;
				case 2 : // Bind -> come back
					if (this.currentTerm != null) {
						if (!this.currentTerm.isEndOfScope() && !this.currentTerm.isLeaf())
							throw new IncorrectTermException(this.currentTerm);
						this.nextTerm = this.currentTerm.getParent();
					}
					this.isWalkingForward = false;
					break;
				default :
					throw new IllegalStateException();
			}
		}
		else if (operator.isDelimiter()) {
			int index = ((Delimiter) operator).getIndex();
			if (this.currentPort.isPrincipal()) {
				if (index == 0) {
					this.walkStack.push(
						this.isWalkingForward
							? operatorFactory.createWalk(0)
							: this.walkUnStack.pop());
				}
				else {
					index--;
					Walk walk = this.walkStack.getFromTop(index).getWalkStack().pop();
					this.walkStack.insertFromTop(index, walk);
				}
				this.currentPort = operator.getPort1();
			}
			else {
				if (index == 0) {
					Walk walk = this.walkStack.pop();
					if (this.isWalkingForward)
						this.walkUnStack.push(walk);
					else
						walk.returnToFactory();
				}
				else {
					index--;
					Walk walk = this.walkStack.removeFromTop(index);
					this.walkStack.getFromTop(index).getWalkStack().push(walk);
				}
				this.currentPort = operator.getPort0();
			}
		}
		else if (operator.isDuplicator()) {
			int index = ((Duplicator) operator).getIndex();
			if (index >= this.walkStack.getSize())
				throw new BadFormedException(this.currentPort);
			Walk walk = this.walkStack.getFromTop(index);
			if (this.currentPort.isPrincipal()) {
				int number = walk.popDirector();
				this.currentPort = operator.getPort(number);
			}
			else {
				walk.pushDirector(this.currentPort.getNumber());
				this.currentPort = operator.getPort0();
			}
		}
		else if (operator.isEraser()) {
			if (operator != this.rootEraser)
				throw new BadFormedException(this.currentPort);
			this.walkStack.returnToFactory();
			this.walkStack = null;
			this.walkUnStack.returnToFactory();
			this.walkUnStack = null;
			this.isWalkingForward = true;
			this.currentTerm = null;
			this.nextTerm = this.rootTerm;
		}
		else
			throw new IllegalStateException();
	}

	public boolean isReturnedToRoot() {
		return this.walkStack == null;
	}

	public void returnToRoot() {
		if (this.walkStack == null)
			return;
		this.walkStack.returnToFactory();
		this.walkStack = null;
		this.walkUnStack.returnToFactory();
		this.walkUnStack = null;
		this.bodyIndex = 0;
		this.currentPort = this.rootEraser.getPort();
		this.isWalkingForward = true;
	}

	public void printTour(PrintStream out) {
		while (true) {
			out.println(this);
			this.executeStep();
			if (this.isReturnedToRoot())
				break;
		}
	}

	//////////////////////////////////////////
	// WalkingContextTranslater implementation
	//////////////////////////////////////////

	public void setRootTerm(Term rootTerm) {
		if (this.walkStack != null)
			throw new IllegalStateException();
		this.rootTerm = rootTerm;
		this.nextTerm = rootTerm;
	}

	public Term getCurrentTerm() {
		Operator operator = this.currentPort.getOperator();
		return (operator.isAbstractor() || operator.isApplicator())
			? this.currentTerm
			: null;
	}
}
