/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.generate;

import java.io.PrintStream;

import org.xmloperator.lambda.tree.exception.MissingContextException;
import org.xmloperator.lambda.tree.exception.NotInNormalFormException;
import org.xmloperator.lambda.tree.exception.TermException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.reduction.TreeBetaUtils;
import org.xmloperator.lambda.tree.reduction.ReductionCounter;
import org.xmloperator.lambda.tree.reduction.ReductionTracer;
import org.xmloperator.lambda.tree.serialize.SerializeUtils;
import org.xmloperator.lambda.tree.serialize.Serializer;
import org.xmloperator.lambda.tree.util.AlphaUtils;
import org.xmloperator.lambda.tree.util.FactoryUtils;
import org.xmloperator.lambda.tree.util.SignatureUtils;
import org.xmloperator.lambda.tree.util.SubstitutionUtils;
import org.xmloperator.lambda.tree.util.WellFormedness;

/**
 * Generator of all the Terms having a given tree size.
 */
public class FullGenerator implements TermIterator {
	private static final boolean ASSERTION = Test.ASSERTION;
	private static final boolean TRACE = false;

	private final TermFactory termFactory;
	private final int treeSize;
	private final boolean closedTermsOnly;
	private Term term;

	public FullGenerator(
		TermFactory termFactory,
		int treeSize,
		boolean closedTermsOnly) {
		if (treeSize < 2)
			throw new IllegalArgumentException();
		this.termFactory = termFactory;
		this.treeSize = treeSize;
		this.closedTermsOnly = closedTermsOnly;
		this.term = firstTerm(termFactory, treeSize);
		if (ASSERTION) {
			WellFormedness.verifyWellFormedness(this.term);
			if (SignatureUtils.leafAndAbstractionCount(this.term) != treeSize)
				throw new TermException(this.term);
		}
	}

	private static final Term firstTerm(
		final TermFactory termFactory,
		final int treeSize) {
		if (treeSize < 1)
			throw new IllegalArgumentException();
		Term term = FactoryUtils.newLeaf(termFactory, treeSize - 2);
		for (int k = 0; k < treeSize - 1; k++)
			term = FactoryUtils.newAbstraction(term);
		return term;
	}

	private static final Term nextEquiSizedTerm(Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			int abstractionCount = 0;
			int lostSize = 0;
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf()) {
					SubstitutionUtils.remove(term);
					term.returnToFactory();
					term = null;
					lostSize++;
				}
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					SubstitutionUtils.removeTermWithoutItsBody(endOfScope);
					endOfScope.returnToFactory();
					return rootTerm;
				}
				else if (term.isAbstraction()) {
					abstractionCount++;
					Abstraction abstraction = (Abstraction) term;
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getArgument();
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty())
						return null;
					term = termStack.pop();
					if (term.isAbstraction()) {
						lostSize++;
						if (lostSize >= 4 || (lostSize >= 2 && abstractionCount != 0)) {
							Application application = termFactory.createApplication();
							Term argument =
								abstractionCount == 0
									? (Term) FactoryUtils.newAbstraction(
										FactoryUtils.newLeaf(termFactory, abstractionCount))
									: FactoryUtils.newLeaf(termFactory, abstractionCount - 1);
							if (argument.isAbstraction())
								lostSize -= 2;
							else
								lostSize--;
							application.setArgument(argument);
							application.setBody(firstTerm(termFactory, lostSize));
							if (term == rootTerm)
								rootTerm = application;
							else
								SubstitutionUtils.substitute(term, application);
							term.returnToFactory();
							return rootTerm;
						}
						SubstitutionUtils.remove(term);
						term.returnToFactory();
						abstractionCount--;
					}
					else if (term.isApplication()) {
						Application application = (Application) term;
						if (application.getArgument() == null) {
							application.setArgument(firstTerm(termFactory, lostSize));
							lostSize = 0;
							termStack.push(application);
							term = application.getBody();
							break;
						}
						else if (application.getBody() == null) {
							Term argument = application.getArgument();
							application.clearArgument();
							if (lostSize > 2 || (lostSize == 2 && abstractionCount != 0)) {
								application.setArgument(FactoryUtils.newAbstraction(argument));
								application.setBody(firstTerm(termFactory, lostSize - 1));
								return rootTerm;
							}
							lostSize += returnTerm(argument);
							SubstitutionUtils.remove(application);
							application.returnToFactory();
						}
						else
							throw new IllegalStateException();
					}
					else
						throw new IllegalStateException();
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	private static final int parentEndOfScopeCount(final Term leaf) {
		int count = 0;
		Term parent = leaf.getParent();
		while (parent != null && parent.isEndOfScope()) {
			count++;
			parent = parent.getParent();
		}
		return count;
	}

	private static final int returnTerm(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			int lostSize = 0;
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf()) {
					term.returnToFactory();
					lostSize++;
				}
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					term = endOfScope.getBody();
					endOfScope.clearBody();
					endOfScope.returnToFactory();
					continue;
				}
				else if (term.isAbstraction()) {
					Abstraction abstraction = (Abstraction) term;
					term = abstraction.getBody();
					abstraction.clearBody();
					abstraction.returnToFactory();
					lostSize++;
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getArgument();
					application.clearArgument();
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					return lostSize;
				Application application = (Application) termStack.pop();
				term = application.getBody();
				application.clearBody();
				application.returnToFactory();
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final Serializer SERIALIZER = Test.SERIALIZER;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Executes some Term generations.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= TRACE | IS_VERBOSE;
			boolean isOk = true;
			int maxTreeSize = 7;
			for (int k = 2; k <= maxTreeSize; k++)
				isOk &= test(out, isVerbose, k);
			return isOk;
		}
		catch (TermException exception) {
			System.out.println("FullTermGenerator TermException");
			Term subTerm = exception.getTerm();
			if (SERIALIZER != null && subTerm != null)
				SerializeUtils.printUnderlinedTerm(SERIALIZER, null, subTerm);
			exception.printStackTrace(out);
			return false;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static final boolean test(
		PrintStream out,
		boolean isVerbose,
		int treeSize) {
		TermFactory termFactory = TERM_FACTORY;
		TermIterator iterator = new FullGenerator(termFactory, treeSize, true);
		int count = 0;
		int normalCount = 0;
		int newCount = 0;
		while (iterator.hasNext()) {
			count++;
			Term term = iterator.next();
			if (TreeBetaUtils.leftermostOutermostRedex(term) == null)
				normalCount++;
			else {
				if (isNewTerm(treeSize, term)) {
					newCount++;
					if (TRACE) {
						System.out.print("---");
						SerializeUtils.printTerm(SERIALIZER, term);
					}
					ReductionCounter reductionCounter = new ReductionCounter(5);
					if (TRACE)
						reductionCounter.nextReductionListener =
							new ReductionTracer(SERIALIZER, false, false, term);
					try {
						term = TreeBetaUtils.reduceDeeply(term, reductionCounter);
					}
					catch (NotInNormalFormException exception) {
						term = exception.getTerm();
					}
				}
			}
			FactoryUtils.returnTerm(term);
		}
		if (isVerbose)
			System.out.println(
				"FullGenerator("
					+ treeSize
					+ ") : "
					+ count
					+ " term(s), "
					+ normalCount
					+ " normal, "
					+ newCount
					+ " new.");
		return true;
	}

	private static final boolean isNewTerm(int treeSize, Term term) {
		int iterationCount = 5;
		Term cloneTerm = AlphaUtils.clone(term);
		try {
			for (int k = 0; k < iterationCount; k++) {
				Application mainRedex = TreeBetaUtils.leftermostOutermostRedex(cloneTerm);
				if (mainRedex == null)
					return false;
				boolean isRoot = mainRedex == cloneTerm;
				Term result = TreeBetaUtils.reduce(mainRedex);
				if (isRoot)
					cloneTerm = result;
				if (SignatureUtils.leafAndAbstractionCount(cloneTerm) < treeSize)
					return false;
			}
			return true;
		}
		finally {
			FactoryUtils.returnTerm(cloneTerm);
		}
	}

	////////////////////////////////////
	// TermIterator implementation
	////////////////////////////////////

	public boolean hasNext() {
		return this.term != null;
	}

	public Term next() {
		if (this.term == null)
			return null;
		Term cloneTerm = AlphaUtils.clone(this.term);
		while (true)
			try {
				this.term = nextEquiSizedTerm(this.term);
				if (this.term != null) {
					if (closedTermsOnly)
						AlphaUtils.verifyClosure(this.term);
					if (ASSERTION) {
						WellFormedness.verifyWellFormedness(this.term);
						if (SignatureUtils.leafAndAbstractionCount(this.term)
							!= this.treeSize)
							throw new TermException(this.term);
					}
				}
				break;
			}
			catch (MissingContextException exception) {
			}
		return cloneTerm;
	}
}
