/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.util;

import java.io.PrintStream;

import org.xmloperator.lambda.tree.exception.MissingContextException;
import org.xmloperator.lambda.tree.exception.UnextrudedEndOfScopeException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.model.TermWithBody;

/**
 * Utilities about cloning, scope extrusion and alpha-equality.
 */
public abstract class AlphaUtils {
	private static final boolean ASSERTION = Test.ASSERTION;
	
	/**
	 * Returns the Abstraction matching a given variable.
	 * Returns null if there is the variable is free.
	 * @param variable a Leaf or an EndOfScope.
	 * @return the Abstraction matching a given variable, if any, null elsewhere.
	 */
	public static final Abstraction matchingAbstraction(final Term variable) {
		if (!variable.isLeaf() && !variable.isEndOfScope())
			throw new IllegalArgumentException();
		int endOfScopeCount = 0;
		Term parent = variable.getParent();
		while (parent != null) {
			if (parent.isEndOfScope())
				endOfScopeCount++;
			else if (parent.isAbstraction()) {
				if (endOfScopeCount == 0)
					return (Abstraction) parent;
				endOfScopeCount--;
			}
			parent = parent.getParent();
		}
		return null;
	}

	/**
	 * Returns the index of an <tt>Abstraction</tt> within a Term. The first
	 *   Abstraction in the Term has <tt>0</tt> as index. Ignores the EndOfSCopes.
	 * @param rootTerm a Term containing the given Abstraction.
	 * @param abstraction an Abstraction.
	 * @return the index of an <tt>Abstraction</tt> within a Term.
	 */
	public static final int abstractionIndex(
		final Term rootTerm,
		final Abstraction abstraction) {
		int index = 0;
		Term term = abstraction;
		while (term != null && term != rootTerm) {
			term = term.getParent();
			if (term.isAbstraction())
				index++;
		}
		return index;
	}

	/**
	 * Verify that a Term is closed, i.e. there is no free variable in it.
	 * @param rootTerm a Term.
	 * @throws MissingContextException if the given Term is not closed. The free
	 *   variable is provided by the exception.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final void verifyClosure(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			int abstractionStackSize = 0;
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf()) {
					if (abstractionStackSize == 0)
						throw new MissingContextException(term);
				}
				else if (term.isEndOfScope()) {
					if (abstractionStackSize == 0)
						throw new MissingContextException(term);
					abstractionStackSize--;
					EndOfScope endOfScope = (EndOfScope) term;
					termStack.push(endOfScope);
					term = endOfScope.getBody();
					continue;
				}
				else if (term.isAbstraction()) {
					abstractionStackSize++;
					Abstraction abstraction = (Abstraction) term;
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty()) {
						term = null;
						break;
					}
					term = termStack.pop();
					if (term.isEndOfScope())
						abstractionStackSize++;
					else if (term.isAbstraction())
						abstractionStackSize--;
					else if (term.isApplication()) {
						Application application = (Application) term;
						term = application.getArgument();
						break;
					}
					else
						throw new IllegalStateException();
				}
				if (term == null)
					break;
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	/**
	 * Evaluates the strict-equivalence of two Terms.
	 * @param term1 a first Term.
	 * @param term2 a second Term.
	 * @return true if the two given Term are strictly equivalent.
	 * @throws IllegalArgumentException if some argument is null.
	 * @throws NullPointerException if some argument term is broken.
	 */
	public static final boolean strictlyEquals(
		final Term rootTerm1,
		final Term rootTerm2) {
		TermFactory termFactory = rootTerm1.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term1 = rootTerm1;
			Term term2 = rootTerm2;
			while (true) {
				if (term1.isLeaf() || term2.isLeaf()) {
					if (!term1.isLeaf() || !term2.isLeaf())
						return false;
				}
				else if (term1.isEndOfScope() || term2.isEndOfScope()) {
					if (!term1.isEndOfScope() || !term2.isEndOfScope())
						return false;
					EndOfScope endOfScope1 = (EndOfScope) term1;
					EndOfScope endOfScope2 = (EndOfScope) term2;
					term1 = endOfScope1.getBody();
					term2 = endOfScope2.getBody();
					continue;
				}
				else if (term1.isAbstraction() || term2.isAbstraction()) {
					if (!term1.isAbstraction() || !term2.isAbstraction())
						return false;
					Abstraction abstraction1 = (Abstraction) term1;
					Abstraction abstraction2 = (Abstraction) term2;
					term1 = abstraction1.getBody();
					term2 = abstraction2.getBody();
					continue;
				}
				else if (term1.isApplication() || term2.isApplication()) {
					if (!term1.isApplication() || !term2.isApplication())
						return false;
					Application application1 = (Application) term1;
					Application application2 = (Application) term2;
					termStack.push(application1.getArgument());
					termStack.push(application2.getArgument());
					term1 = application1.getBody();
					term2 = application2.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					break;
				term2 = termStack.pop();
				term1 = termStack.pop();
			}
			return true;
		}
		finally {
			termStack.returnToFactory();
		}
	}

	/**
	 * Evaluates the alpha-equivalence of two closed Terms.
	 * @param term1 a first Term.
	 * @param term2 a second Term.
	 * @return true if the two given Term are alpha-equivalent.
	 * @throws IllegalArgumentException if some argument is null.
	 * @throws NullPointerException if some argument term is broken.
	 * @throws MissingContextException if one of the given Term is not closed.
	 */
	public static final boolean alphaEquals(
		final Term rootTerm1,
		final Term rootTerm2) {
		TermFactory termFactory = rootTerm1.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term1 = rootTerm1;
			Term term2 = rootTerm2;
			while (true) {
				if (term1.isEndOfScope()) {
					EndOfScope endOfScope1 = (EndOfScope) term1;
					term1 = endOfScope1.getBody();
					continue;
				}
				if (term2.isEndOfScope()) {
					EndOfScope endOfScope2 = (EndOfScope) term2;
					term2 = endOfScope2.getBody();
					continue;
				}
				if (term1.isLeaf() || term2.isLeaf()) {
					if (!term1.isLeaf() || !term2.isLeaf())
						return false;
					Abstraction abstraction1 = matchingAbstraction(term1);
					if (abstraction1 == null)
						throw new MissingContextException(term1);
					Abstraction abstraction2 = matchingAbstraction(term2);
					if (abstraction2 == null)
						throw new MissingContextException(term2);
					if (abstractionIndex(rootTerm1, abstraction1)
						!= abstractionIndex(rootTerm2, abstraction2))
						return false;
				}
				else if (term1.isAbstraction() || term2.isAbstraction()) {
					if (!term1.isAbstraction() || !term2.isAbstraction())
						return false;
					Abstraction abstraction1 = (Abstraction) term1;
					Abstraction abstraction2 = (Abstraction) term2;
					term1 = abstraction1.getBody();
					term2 = abstraction2.getBody();
					continue;
				}
				else if (term1.isApplication() || term2.isApplication()) {
					if (!term1.isApplication() || !term2.isApplication())
						return false;
					Application application1 = (Application) term1;
					Application application2 = (Application) term2;
					termStack.push(application1.getArgument());
					termStack.push(application2.getArgument());
					term1 = application1.getBody();
					term2 = application2.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					break;
				term2 = termStack.pop();
				term1 = termStack.pop();
			}
			return true;
		}
		finally {
			termStack.returnToFactory();
		}
	}

	private static final Term cloneNode(Term term) {
		TermFactory termFactory = term.getFactory();
		if (term.isLeaf())
			return termFactory.createLeaf();
		if (term.isEndOfScope())
			return termFactory.createEndOfScope();
		if (term.isAbstraction())
			return termFactory.createAbstraction();
		if (term.isApplication())
			return termFactory.createApplication();
		throw new IllegalArgumentException();
	}

	private static final void attachNode(TermWithBody parent, Term child) {
		if (parent.isEndOfScope()) {
			EndOfScope endOfScope = (EndOfScope) parent;
			endOfScope.setBody(child);
		}
		else if (parent.isAbstraction()) {
			Abstraction abstraction = (Abstraction) parent;
			abstraction.setBody(child);
		}
		else if (parent.isApplication()) {
			Application application = (Application) parent;
			if (application.getBody() == null)
				application.setBody(child);
			else
				application.setArgument(child);
		}
		else
			throw new IllegalArgumentException();
	}

	/**
	 * Clones a Term.
	 * @param rootTerm a Term.
	 * @return the created Term.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final Term clone(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			Term cloneRootTerm = null;
			TermWithBody cloneParent = null;
			while (true) {
				Term cloneTerm = cloneNode(term);
				if (cloneRootTerm == null)
					cloneRootTerm = cloneTerm;
				else
					attachNode(cloneParent, cloneTerm);
				if (!term.isLeaf()) {
					if (term.isApplication()) {
						termStack.push(term);
						termStack.push(cloneTerm);
					}
					cloneParent = (TermWithBody) cloneTerm;
					term = ((TermWithBody) term).getBody();
					continue;
				}
				if (termStack.isEmpty())
					break;
				cloneParent = (Application) termStack.pop();
				Application application = (Application) termStack.pop();
				term = application.getArgument();
			}
			if (ASSERTION && !strictlyEquals(rootTerm, cloneRootTerm))
				throw new IllegalStateException();
			return cloneRootTerm;
		}
		finally {
			termStack.returnToFactory();
		}
	}

	/**
	 * Verify that all the EndOfScopes of a Term are extruded.
	 * There must be no EndOfScope whose child is an Abstraction or an
	 *   Application.
	 * @param rootTerm a Term.
	 * @throws UnextrudedEndOfScopeException if there is some EndOfScope
	 *   whose child is an Abstraction or an Application.
	 */
	public static final void verifyScopeExtrusion(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf());
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					term = endOfScope.getBody();
					if (term.isAbstraction() || term.isApplication())
						throw new UnextrudedEndOfScopeException();
					continue;
				}
				else if (term.isAbstraction()) {
					Abstraction abstraction = (Abstraction) term;
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application.getArgument());
					term = application.getArgument();
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					break;
				term = termStack.pop();
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}
	
	/**
	 * Create a normalized clone, whose all internal EndOfScopes are extruded.
	 * @param rootTerm a Term.
	 * @return the created normalized Term.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final Term normalizedClone(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			Term cloneRootTerm = null;
			TermWithBody cloneParent = null;
			int abstractionStackSize = 0;
			while (true) {
				if (term.isEndOfScope()) {
					term = ((EndOfScope) term).getBody();
					continue;
				}
				if (term.isLeaf()) {
					Abstraction abstraction = matchingAbstraction(term);
					if (abstraction == null)
						throw new MissingContextException(term);
					int endOfScopeCount =
						abstractionStackSize - abstractionIndex(rootTerm, abstraction);
					for (int k = 0; k < endOfScopeCount; k++) {
						EndOfScope endOfScope = termFactory.createEndOfScope();
						attachNode(cloneParent, endOfScope);
						cloneParent = endOfScope;
					}
				}
				Term cloneTerm = cloneNode(term);
				if (cloneRootTerm == null)
					cloneRootTerm = cloneTerm;
				else
					attachNode(cloneParent, cloneTerm);
				if (!term.isLeaf()) {
					if (term.isApplication()) {
						termStack.push(term);
						termStack.push(cloneTerm);
					}
					cloneParent = (TermWithBody) cloneTerm;
					term = ((TermWithBody) term).getBody();
					continue;
				}
				if (termStack.isEmpty())
					break;
				cloneParent = (Application) termStack.pop();
				Application application = (Application) termStack.pop();
				term = application.getArgument();
			}
			if (ASSERTION)
				verifyScopeExtrusion(cloneRootTerm);
			return cloneRootTerm;
		}
		finally {
			termStack.returnToFactory();
		}
	}

	/**
	 * Extrudes an EndOfScope completely.
	 * Does nothing if the body of the given <tt>EndOfScope</tt> is neither
	 *   an <tt>Abstraction</tt> nor an <tt>Application</tt>.
	 * @param endOfScope an EndOfScope.
	 */
	public static final void extrude(final EndOfScope endOfScope) {
		Term body = endOfScope.getBody();
		if (!body.isAbstraction() && !body.isApplication())
			return;
		TermFactory termFactory = endOfScope.getFactory();
		SubstitutionUtils.removeTermWithoutItsBody(endOfScope);
		endOfScope.returnToFactory();
		TermStack termStack = termFactory.createTermStack(false);
		int contextSize = 0;
		try {
			while (true) {
				if (body.isAbstraction()) {
					contextSize++;
					termStack.push(body);
					body = ((Abstraction) body).getBody();
					continue;
				}
				else if (body.isApplication()) {
					termStack.push(body);
					body = ((Application) body).getBody();
					continue;
				}
				else {
					if (contextSize == 0) {
						EndOfScope newEndOfScope = termFactory.createEndOfScope();
						SubstitutionUtils.insert(newEndOfScope, body);
					}
					else {
						if (body.isEndOfScope()) {
							contextSize--;
							termStack.push(body);
							body = ((EndOfScope) body).getBody();
							continue;
						}
					}
				}
				while (true) {
					if (termStack.isEmpty()) {
						body = null;
						break;
					}
					body = termStack.pop();
					if (body.isEndOfScope())
						contextSize++;
					else if (body.isAbstraction())
						contextSize--;
					else if (body.isApplication()) {
						body = ((Application) body).getArgument();
						break;
					}
					else
						throw new IllegalStateException();
				}
				if (body == null)
					break;
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Verify equality and unequality between some terms.
	 * @param out the PrintStream to print results.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			TermFactory termFactory = TERM_FACTORY;
			boolean isOk = true;
			{
				Abstraction i1 = CommonTerms.newIdentity(termFactory);
				Abstraction i2 = (Abstraction) clone(i1);
				isOk &= test(out, isVerbose, i1, i2, "I", "I", true, false);
				FactoryUtils.returnTerm(i1);
				FactoryUtils.returnTerm(i2);
			}
			{
				Abstraction k1 = CommonTerms.newBoolean(termFactory, true);
				Abstraction f1 = CommonTerms.newBoolean(termFactory, false);
				isOk &= test(out, isVerbose, k1, f1, "K", "F", false, false);
				Abstraction k2 = (Abstraction) clone(k1);
				isOk &= test(out, isVerbose, k1, k2, "K", "K", true, false);
				Abstraction f2 = (Abstraction) clone(f1);
				isOk &= test(out, isVerbose, f1, f2, "F", "F", true, false);
				FactoryUtils.returnTerm(k1);
				FactoryUtils.returnTerm(f1);
				FactoryUtils.returnTerm(k2);
				FactoryUtils.returnTerm(f2);
			}
			{
				Abstraction f1 =
					CommonTerms.newNullAbstraction(CommonTerms.newIdentity(termFactory));
				Abstraction f2 = (Abstraction) normalizedClone(f1);
				Abstraction f3 =
					FactoryUtils.newAbstraction(CommonTerms.newIdentity(termFactory));
				isOk &= test(out, isVerbose, f2, f3, "aeav", "aav", true, true);
				FactoryUtils.returnTerm(f1);
				FactoryUtils.returnTerm(f2);
				FactoryUtils.returnTerm(f3);
			}
			{
				Abstraction f1 =
					CommonTerms.newNullAbstraction(CommonTerms.newIdentity(termFactory));
				extrude((EndOfScope) f1.getBody());
				Abstraction f2 =
					FactoryUtils.newAbstraction(CommonTerms.newIdentity(termFactory));
				isOk &= test(out, isVerbose, f1, f2, "aeav", "aav", true, true);
				FactoryUtils.returnTerm(f1);
				FactoryUtils.returnTerm(f2);
			}
			{
				Application omega1 = CommonTerms.newOmega(termFactory);
				Application omega2 = (Application) clone(omega1);
				isOk
					&= test(out, isVerbose, omega1, omega2, "Omega", "Omega", true, false);
				FactoryUtils.returnTerm(omega1);
				FactoryUtils.returnTerm(omega2);
			}
			return isOk;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static boolean test(
		PrintStream out,
		boolean isVerbose,
		Term term1,
		Term term2,
		String name1,
		String name2,
		boolean equals,
		boolean isStrict) {
		boolean isOk =
			(isStrict ? strictlyEquals(term1, term2) : alphaEquals(term1, term2))
				== equals;
		if (isVerbose || !isOk) {
			out.print("AlphaUtils ");
			out.print(isOk ? "OK" : "KO");
			out.println(" (" + name1 + (equals ? " == " : " != ") + name2 + ")");
		}
		return isOk;
	}
}
