/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.util;

import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermWithBody;

/**
 * Utilities for substitutions.
 */
public abstract class SubstitutionUtils {

	/**
	 * Replace a destination Term by a source Term.
	 * If destination has a parent then source receives this parent.
	 * If destination has no parent then do nothing.
	 * @param destination a destination Term.
	 * @param source a source Term.
	 * @throws IllegalArgumentException if the source Term is null or has
	 *   a parent.
	 * @throws NullPOinterException if the source or the destination Term is
	 *   broken.
	 */
	public static final void substitute(Term destination, Term source) {
		Term parent = destination.getParent();
		if (parent == null)
			return;
		if (parent.isEndOfScope() || parent.isAbstraction())
			((TermWithBody)parent).setBody(source);
		else if (parent.isApplication()) {
			Application application = (Application) parent;
			if (destination == application.getBody())
				application.setBody(source);
			else if (destination == application.getArgument())
				application.setArgument(source);
			else
				throw new IllegalArgumentException();
		}
		else
			throw new IllegalArgumentException();
	}

	/**
	 * Removes a Term with its content, if any.
	 * If the given Term has no parent then do nothing.
	 * @param term a Term to remove.
	 */
	public static final void remove(Term term) {
		Term parent = term.getParent();
		if (parent == null)
			return;
		if (parent.isEndOfScope() || parent.isAbstraction())
			((TermWithBody)parent).clearBody();
		else if (parent.isApplication()) {
			Application application = (Application) parent;
			if (term == application.getBody())
				application.clearBody();
			else if (term == application.getArgument())
				application.clearArgument();
			else
				throw new IllegalArgumentException();
		}
		else
			throw new IllegalArgumentException();
	}

	/**
	 * Inserts a <tt>TermWithBody</tt> in a parent to body chain.
	 * @param termWithBody a TermWithBody for inserting as parent of the body.
	 * @param body a Term for becoming the body.
	 * @throws IllegalArgumentException if the given TermWithBody has already
	 *   a parent or a body or if the given body is null.
	 */
	public static final void insert(TermWithBody termWithBody, Term body) {
		if (termWithBody.getBody() != null)
			throw new IllegalArgumentException();
		substitute(body, termWithBody);
		termWithBody.setBody(body);
	}

	/**
	 * Removes a TermWithBody from a parent to body chain.
	 * @param termWithBody a TermWithBody to remove.
	 * @return the body.
	 * @throws NullPointerException if the given TermWithBody is broken.
	 */
	public static final Term removeTermWithoutItsBody(TermWithBody termWithBody) {
		Term body = termWithBody.getBody();
		termWithBody.clearBody();
		substitute(termWithBody, body);
		return body;
	}
}
